# flake8: noqa
# Copy pasted from
# https://raw.githubusercontent.com/greyside/django-bitfield/04efae91d292bffb569bd7cc13e5c941e42e8698/bitfield/forms.py
from bitfield.types import BitHandler
from django.forms import CheckboxSelectMultiple
from django.forms import IntegerField
from django.forms import ValidationError
from django.utils.safestring import mark_safe

try:
    from django.utils.encoding import force_text
except ImportError:
    from django.utils.encoding import force_unicode as force_text


class CheckboxSelectMultipleP(CheckboxSelectMultiple):
    # https://djangosnippets.org/snippets/1760/

    def render(self, *args, **kwargs):
        output = super().render(*args, **kwargs)
        return (mark_safe(output.replace(u'<ul>', u'').
                          replace(u'</ul>', u'').
                          replace(u'<li>', u'<p>').
                          replace(u'</li>', u'</p>')))


class BitFieldCheckboxSelectMultiple(CheckboxSelectMultipleP):

    def render(self, name, value, attrs=None, renderer=None):
        if isinstance(value, BitHandler):
            value = [k for k, v in value if v]
        return super().render(name, value, attrs=attrs)

    def _has_changed(self, initial, data):
        if initial is None:
            initial = []
        if data is None:
            data = []
        if initial != data:
            return True
        initial_set = set([force_text(value) for value in initial])
        data_set = set([force_text(value) for value in data])
        return data_set != initial_set


class BitFormField(IntegerField):

    def __init__(self,
                 choices=(),
                 widget_choices=(),
                 widget=BitFieldCheckboxSelectMultiple,
                 *args,
                 **kwargs):
        initial = kwargs.get('initial')
        if isinstance(initial, int):
            l = []
            for i in range(0, 63):
                if (1 << i) & initial > 0:
                    l += [choices[i][0]]
            kwargs['initial'] = l
        self.widget = widget
        super(BitFormField, self).__init__(widget=widget, *args, **kwargs)
        self.choices = choices
        # widget choices is a subset of choices that are available
        self.widget.choices = widget_choices or choices

    def clean(self, value):
        if not value:
            return 0

        # Assume an iterable which contains an item per flag that's enabled
        result = BitHandler(0, [k for k, v in self.choices])
        for k in value:
            try:
                setattr(result, str(k), True)
            except AttributeError:
                raise ValidationError('Unknown choice: %r' % (k,))
        return int(result)



from django.core.management.base import BaseCommand

from orchestra.models import Todo

import logging
logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = ('Migrates completion and skipped_datetime'
            ' values to status field in Todo model.')

    def bulk_update_todo_status(self, queryset, status):
        num_rows = queryset.count()
        logger.info('Started updating %s todos with status = %s',
                    num_rows, status)
        queryset.only('id', 'status').update(status=status)
        logger.info('Finished updating %s todos with status = %s',
                    num_rows, status)

    def handle(self, *args, **options):
        # declined todos
        self.bulk_update_todo_status(
            queryset=Todo.objects.filter(
                skipped_datetime__isnull=False),
            status=Todo.Status.DECLINED.value
        )
        # completed todos
        self.bulk_update_todo_status(
            queryset=Todo.objects.filter(
                skipped_datetime__isnull=True,
                completed=True),
            status=Todo.Status.COMPLETED.value
        )
        # pending todos
        self.bulk_update_todo_status(
            queryset=Todo.objects.filter(
                skipped_datetime__isnull=True,
                completed=False),
            status=Todo.Status.PENDING.value
        )



# Generated by Django 2.2.13 on 2020-08-18 16:28

from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0085_rename_todo_description_to_title'),
    ]

    operations = [
        migrations.AddField(
            model_name='todo',
            name='additional_data',
            field=jsonfield.fields.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name='todo',
            name='details',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='todo',
            name='order',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='todo',
            name='project',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='todos', to='orchestra.Project'),
        ),
        migrations.AddField(
            model_name='todo',
            name='section',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='todo',
            name='status',
            field=models.IntegerField(blank=True, choices=[(0, 'pending'), (1, 'completed'), (2, 'declined')], null=True),
        ),
        migrations.AddField(
            model_name='todo',
            name='step',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='todos', to='orchestra.Step'),
        ),
        migrations.AlterField(
            model_name='todo',
            name='task',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='todos', to='orchestra.Task'),
        ),
    ]



import datetime
from rest_framework import serializers

from orchestra.models import Iteration
from orchestra.models import Project
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.models import TaskTimer
from orchestra.models import TimeEntry
from orchestra.models import WorkerCertification


class ProjectSerializer(serializers.ModelSerializer):

    class Meta:
        model = Project
        fields = (
            'id',
            'workflow_slug',
            'workflow_version_slug',
            'short_description',
            'start_datetime',
            'priority',
            'project_data',
            'scratchpad_url',
            'task_class',
            'status',
            'slack_group_id'
        )

    workflow_slug = serializers.SerializerMethodField()

    def get_workflow_slug(self, obj):
        return obj.workflow_version.workflow.slug

    workflow_version_slug = serializers.SlugRelatedField(
        source='workflow_version', slug_field='slug', read_only=True)

    task_class = serializers.ChoiceField(
        choices=WorkerCertification.TASK_CLASS_CHOICES)

    project_data = serializers.SerializerMethodField()

    def get_project_data(self, obj):
        """
        This function exists to automatically deserialize the JSON blob from
        the `project_data` JSONField
        """
        return obj.project_data


class ProjectSummarySerializer(serializers.ModelSerializer):

    class Meta:
        model = Project
        fields = (
            'id',
            'short_description',
            'start_datetime',
            'status'
        )

    status = serializers.SerializerMethodField()

    def get_status(self, obj):
        return dict(Project.STATUS_CHOICES)[obj.status]


class TaskSerializer(serializers.ModelSerializer):

    class Meta:
        model = Task
        fields = (
            'id',
            'step_slug',
            'project',
            'status',
            'latest_data',
            'assignments',
            'start_datetime'
        )

    step_slug = serializers.SlugRelatedField(source='step',
                                             slug_field='slug',
                                             read_only=True)

    status = serializers.SerializerMethodField()
    latest_data = serializers.SerializerMethodField()
    assignments = serializers.SerializerMethodField()

    def get_status(self, obj):
        return dict(Task.STATUS_CHOICES).get(obj.status, None)

    def get_latest_data(self, obj):
        """
        Return latest input data for a specified task.

        Args:
            task (orchestra.models.Task):
                The task object for which to retrieve data.

        Returns:
            latest_data (str):
                A serialized JSON blob containing the latest input data.
        """
        active_assignment = (obj.assignments
                             .filter(status=TaskAssignment.Status.PROCESSING))
        if active_assignment.exists():
            assignment = active_assignment[0]
        else:
            assignment = (obj.assignments.all()
                          .order_by('-assignment_counter').first())
        if not assignment:
            return None

        latest_data = assignment.in_progress_task_data
        return latest_data

    def get_assignments(self, obj):
        assignments = TaskAssignmentSerializer(obj.assignments.all()
                                               .order_by('assignment_counter'),
                                               many=True)
        return assignments.data


class TaskAssignmentSerializer(serializers.ModelSerializer):
    recorded_work_time = serializers.SerializerMethodField()

    class Meta:
        model = TaskAssignment
        fields = (
            'id',
            'start_datetime',
            'worker',
            'task',
            'status',
            'in_progress_task_data',
            'iterations',
            'recorded_work_time'
        )

    worker = serializers.SerializerMethodField()
    status = serializers.SerializerMethodField()
    in_progress_task_data = serializers.SerializerMethodField()
    iterations = serializers.SerializerMethodField()

    def get_worker(self, obj):
        if not obj.worker:
            return {
                'id': None,
                'username': None,
                'first_name': None,
                'last_name': None,
                'slack_username': None,
                'slack_user_id': None
            }
        return {
            'id': obj.worker.id,
            'username': obj.worker.user.username,
            'first_name': obj.worker.user.first_name,
            'last_name': obj.worker.user.last_name,
            'slack_username': obj.worker.slack_username,
            'slack_user_id': obj.worker.slack_user_id
        }

    def get_status(self, obj):
        return dict(TaskAssignment.STATUS_CHOICES).get(obj.status, None)

    def get_iterations(self, obj):
        iterations = IterationSerializer(
            obj.iterations.order_by('start_datetime'), many=True)
        return iterations.data

    def get_in_progress_task_data(self, obj):
        """
        This function exists to automatically deserialize the JSON blob from
        the `in_progress_task_data` JSONField

        TODO(derek): maybe make a custom JSON serializer field type?
        """
        return obj.in_progress_task_data

    def get_recorded_work_time(self, obj):
        query = TimeEntry.objects.filter(worker=obj.worker, assignment=obj)
        if not query:
            return None

        total_time = datetime.timedelta()
        for t in query:
            total_time += t.time_worked

        return total_time.total_seconds()


class TimeEntrySerializer(serializers.ModelSerializer):

    class Meta:
        model = TimeEntry
        fields = ('id', 'date', 'time_worked', 'description', 'assignment')
        read_only_fields = ('id',)


class TaskTimerSerializer(serializers.ModelSerializer):

    class Meta:
        model = TaskTimer
        fields = '__all__'


class IterationSerializer(serializers.ModelSerializer):

    class Meta:
        model = Iteration
        fields = (
            'id',
            'start_datetime',
            'end_datetime',
            'status',
            'assignment',
            'submitted_data',
        )

    status = serializers.SerializerMethodField()
    submitted_data = serializers.SerializerMethodField()

    def get_status(self, obj):
        return dict(Iteration.STATUS_CHOICES).get(obj.status, None)

    def get_submitted_data(self, obj):
        """
        This function exists to automatically deserialize the JSON blob from
        the `in_progress_task_data` JSONField
        """
        return obj.submitted_data



from django.contrib.auth import get_user_model
from django.urls import reverse
from rest_framework import status

from orchestra.models import Worker
from orchestra.tests.helpers import OrchestraTestCase

UserModel = get_user_model()


class OrchestraRegistrationViewTests(OrchestraTestCase):

    def setUp(self):
        super().setUp()
        self.url = reverse('orchestra:registration_register')

    def get_valid_post_data(self):
        return {
            'username': 'test',
            'email': 'test@test.com',
            'password1': 'test',
            'password2': 'test',  # password confirmation
        }

    def assert_valid_form(self, resp):
        # This asserts that the form is valid since we redirect
        self.assertRedirects(resp, reverse('registration_complete'))

    def assert_related_obj_count(self, count):
        self.assertEqual(UserModel.objects.all().count(), count)
        self.assertEqual(Worker.objects.all().count(), count)

    def assert_models(self, post_data, count=1):
        """
            On a successful register we expect the following:
            1. A User object is created with the form_data we give
            2. A Worker object is associated with this User
        """
        user = UserModel.objects.get(username=post_data.get('username'))
        self.assertEqual(user.email, post_data.get('email'))
        self.assertTrue(user.check_password(post_data.get('password1')))
        self.assertTrue(Worker.objects.filter(user=user).exists())

        self.assert_related_obj_count(count)

    def test_worker_created(self):
        post_data = self.get_valid_post_data()
        resp = self.request_client.post(self.url, post_data)

        self.assert_valid_form(resp)
        self.assert_models(post_data)

    def test_multiple_workers_created(self):
        post_data = self.get_valid_post_data()
        resp = self.request_client.post(self.url, post_data)

        self.assert_valid_form(resp)
        self.assert_models(post_data)

        post_data = self.get_valid_post_data()
        post_data['username'] = 'new_{}'.format(post_data.get('username'))
        post_data['email'] = 'new_{}'.format(post_data.get('email'))
        post_data['password1'] = 'new_{}'.format(post_data.get('password1'))
        post_data['password2'] = 'new_{}'.format(post_data.get('password2'))

        resp = self.request_client.post(self.url, post_data)

        self.assert_valid_form(resp)
        self.assert_models(post_data, count=2)

    def test_invalid_no_worker(self):
        post_data = self.get_valid_post_data()
        post_data['username'] = ''
        self.request_client.post(self.url, post_data)
        self.assert_related_obj_count(0)

    def test_duplicate_register(self):
        post_data = self.get_valid_post_data()

        resp = self.request_client.post(self.url, post_data)

        self.assert_valid_form(resp)
        self.assert_models(post_data)

        resp = self.request_client.post(self.url, post_data)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assert_related_obj_count(1)



class CommunicationPreferenceMixin(object):

    @classmethod
    def get_default_methods(cls):
        """
            We want to set every value in the bitfield to 1.
        """
        return 2 ** len(cls.COMMUNICATION_METHODS) - 1

    def get_descriptions(self):
        key = self.CommunicationType(self.communication_type)
        return self.COMMUNICATION_TYPE_DESCRIPTIONS[key]

    def can_slack(self):
        """
            Boolean of whether or not the Worker wants slack messages
            for the CommunicationType.
        """
        return self.methods.slack

    def can_email(self):
        """
            Boolean of whether or not the Worker wants email messages
            for the CommunicationType.
        """
        return self.methods.email

    def __str__(self):
        return '{} - {} - {}'.format(
            self.worker,
            self.methods,
            self.get_descriptions().get('short_description')
        )


class StaffBotRequestMixin(object):

    def get_request_cause_description(self):
        return self.RequestCause(self.request_cause).description

    def get_request_status_description(self):
        return self.Status(self.status).description

    def __str__(self):
        return '{} - {} - {}'.format(
            self.task,
            self.get_request_cause_description(),
            self.get_request_status_description()
        )


class StaffingRequestInquiryMixin(object):

    def __str__(self):
        return '{} - {}'.format(
            self.communication_preference.worker,
            self.request.task.id
        )


class StaffingResponseMixin(object):

    def __str__(self):
        return '{} - {} - {}'.format(
            self.request_inquiry,
            self.is_available,
            self.is_winner
        )



# Generated by Django 2.2.13 on 2020-08-10 15:03

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0084_workflowversion_abort_completion_function'),
    ]

    operations = [
        migrations.RenameField(
            model_name='todo',
            old_name='description',
            new_name='title',
        ),
    ]



# Generated by Django 2.2.13 on 2020-09-04 14:49


from django.db import migrations
from django.db.models import OuterRef
from django.db.models import Subquery


def fill_project_and_step_fields(apps, schema_editor):
    Step = apps.get_model('orchestra', 'Step')
    Project = apps.get_model('orchestra', 'Project')
    Todo = apps.get_model('orchestra', 'Todo')
    # This is the outer query field the querysets in the subqueries need to refer to
    o = OuterRef('id')
    # Select a step which has a task which has a todo which has an identical ID
    step_outerref_filter_qs = Step.objects.filter(tasks__todos__id=o)
    # Select a project which has a task which has a todo which has an identical ID
    project_outterref_filter_qs = Project.objects.filter(tasks__todos__id=o)
    # We need id value only. These will go as subqueries
    subq_step = step_outerref_filter_qs.values('id')[:1]
    subq_proj = project_outterref_filter_qs.values('id')[:1]
    # Update all todos
    Todo.objects.update(project=Subquery(subq_proj), step=Subquery(subq_step))
    """
    The query above generates the following SQL-query:

    UPDATE "orchestra_todo"
    SET "project_id" = (
          SELECT     u0."id"
          FROM       "orchestra_project" U0
          INNER JOIN "orchestra_task" U1
          ON         (u0."id" = u1."project_id")
          INNER JOIN "orchestra_todo" U2
          ON         (u1."id" = u2."task_id")
          WHERE      u2."id" = ("orchestra_todo"."id") limit 1),
       "step_id" = (
          SELECT     u0."id"
          FROM       "orchestra_step" u0
          INNER JOIN "orchestra_task" u1
          ON         (u0."id" = u1."step_id")
          INNER JOIN "orchestra_todo" u2
          ON         (u1."id" = u2."task_id")
          WHERE      u2."id" = ("orchestra_todo"."id") limit 1)
    WHERE  "orchestra_todo"."is_deleted" = false;args=(false,)
    """

class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0086_add_fields_to_todo'),
    ]

    operations = [
        migrations.RunPython(fill_project_and_step_fields,  # manually-reviewed
                             reverse_code=migrations.RunPython.noop),  # manually-reviewed
    ]



from django.conf.urls import url

from orchestra.interface_api.project_management import views

app_name = 'project_management'

urlpatterns = [
    url(r'^projects/$',
        views.ProjectList.as_view(),
        name='projects'),

    url(r'^assign_task/$',
        views.assign_task_api,
        name='assign_task'),

    url(r'^complete_and_skip_task/$',
        views.complete_and_skip_task_api,
        name='complete_and_skip_task'),

    url(r'^create_subsequent_tasks/$',
        views.create_subsequent_tasks_api,
        name='create_subsequent_tasks'),

    url(r'^edit_slack_membership/$',
        views.edit_slack_membership_api,
        name='edit_slack_membership'),

    url(r'^unarchive_slack_channel/$',
        views.unarchive_slack_channel_api,
        name='unarchive_slack_channel'),

    url(r'^end_project/$',
        views.end_project_api,
        name='end_project'),

    url(r'^set_project_status/$',
        views.set_project_status_api,
        name='set_project_status'),

    url(r'^project_information/$',
        views.project_information_api,
        name='project_information'),

    url(r'^reassign_assignment/$',
        views.reassign_assignment_api,
        name='reassign_assignment'),

    url(r'^revert_task/$',
        views.revert_task_api,
        name='revert_task'),

    url(r'^staff_task/$',
        views.staff_task,
        name='staff_task'),
]



from datetime import timedelta
from unittest.mock import patch

from django.test import override_settings
from django.utils import timezone

from orchestra.bots.errors import StaffingResponseException
from orchestra.communication.staffing import get_available_requests
from orchestra.communication.staffing import handle_staffing_response
from orchestra.communication.staffing import \
    remind_workers_about_available_tasks
from orchestra.communication.staffing import address_staffing_requests
from orchestra.communication.staffing import \
    warn_staffing_team_about_unstaffed_tasks
from orchestra.communication.utils import mark_worker_as_winner
from orchestra.models import CommunicationPreference
from orchestra.models import StaffBotRequest
from orchestra.models import StaffingRequestInquiry
from orchestra.models import StaffingResponse
from orchestra.models import TaskAssignment
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import CertificationFactory
from orchestra.tests.helpers.fixtures import CommunicationPreferenceFactory
from orchestra.tests.helpers.fixtures import StaffBotRequestFactory
from orchestra.tests.helpers.fixtures import StaffingRequestInquiryFactory
from orchestra.tests.helpers.fixtures import StaffingResponseFactory
from orchestra.tests.helpers.fixtures import TimeEntryFactory
from orchestra.tests.helpers.fixtures import WorkerAvailabilityFactory
from orchestra.tests.helpers.fixtures import WorkerCertificationFactory
from orchestra.tests.helpers.fixtures import WorkerFactory


def _get_assignable_hours_for_task(task_data):
    return 3


class StaffingTestCase(OrchestraTestCase):

    def setUp(self):
        self.certification = CertificationFactory()
        self.worker, communication_preference = self._create_worker(0)
        self.staffing_request_inquiry = self._create_inquired_staffing_request(
            self.worker, communication_preference, False)

        super().setUp()

    def _create_worker(self, staffing_priority):
        worker = WorkerFactory()
        WorkerCertificationFactory(
            worker=worker,
            certification=self.certification,
            staffing_priority=staffing_priority
        )
        communication_preference = CommunicationPreferenceFactory(
            worker=worker,
            communication_type=(
                CommunicationPreference.CommunicationType
                .NEW_TASK_AVAILABLE.value))
        return worker, communication_preference

    def _create_inquired_staffing_request(
            self, worker, communication_preference, is_won):
        staffing_request_inquiry = StaffingRequestInquiryFactory(
            communication_preference=communication_preference,
            request__task__step__is_human=True,
            request__status=(
                StaffBotRequest.Status.CLOSED.value
                if is_won else StaffBotRequest.Status.SENDING_INQUIRIES.value)
        )
        staffing_request_inquiry \
            .request.task.step.required_certifications.add(
                self.certification)
        staffing_request = staffing_request_inquiry.request
        staffing_request.task.step.assignable_hours_function = {
            'path': ('orchestra.communication.tests'
                     '.test_staffing._get_assignable_hours_for_task')
        }
        staffing_request.task.step.save()
        if is_won:
            StaffingResponseFactory(
                request_inquiry=staffing_request_inquiry,
                is_available=True,
                is_winner=True)
        return staffing_request_inquiry

    def test_handle_staffing_response_invalid_request(self):
        old_count = StaffingResponse.objects.all().count()

        # Invalid staffing_request_inquiry_id
        response = handle_staffing_response(
            self.worker, None, is_available=True)
        self.assertEqual(response, None)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count)

        # Invalid worker
        response = handle_staffing_response(
            None, self.staffing_request_inquiry.id, is_available=True)
        self.assertEqual(response, None)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count)

    def test_handle_staffing_response_is_available(self):
        # Test StaffingResponse object creation
        old_count = StaffingResponse.objects.all().count()

        # assign task is called
        response = handle_staffing_response(
            self.worker, self.staffing_request_inquiry.id,
            is_available=True)
        self.assertTrue(response.is_winner)
        self.staffing_request_inquiry.refresh_from_db()
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.CLOSED.value)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count + 1)

        task_assignment = (
            TaskAssignment.objects
            .get(worker=self.worker,
                 task=self.staffing_request_inquiry.request.task))
        self.assertEqual(task_assignment.status,
                         TaskAssignment.Status.PROCESSING)
        self.staffing_request_inquiry.request.task.refresh_from_db()

        # Replay of same request
        response = handle_staffing_response(
            self.worker, self.staffing_request_inquiry.id, is_available=True)
        self.assertTrue(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.CLOSED.value)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count + 1)

        # Change mind to `is_available=False` does not do anything
        with self.assertRaises(StaffingResponseException):
            response = handle_staffing_response(
                self.worker, self.staffing_request_inquiry.id,
                is_available=False)

        new_request_inquiry = StaffingRequestInquiryFactory(
            request__task__step__is_human=True
        )
        new_worker = new_request_inquiry.communication_preference.worker
        old_count = StaffingResponse.objects.all().count()
        response = handle_staffing_response(
            new_worker, new_request_inquiry.id, is_available=True)
        self.assertTrue(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.CLOSED.value)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count + 1)

        task_assignment = (
            TaskAssignment.objects
            .get(worker=new_worker,
                 task=new_request_inquiry.request.task))
        self.assertEqual(task_assignment.status,
                         TaskAssignment.Status.PROCESSING)

        # restaff
        response.is_winner = False
        response.save()

        worker2 = WorkerFactory()
        staffing_request_inquiry2 = StaffingRequestInquiryFactory(
            communication_preference__worker=worker2,
            request__task=new_request_inquiry.request.task
        )
        response = handle_staffing_response(
            worker2, staffing_request_inquiry2.id, is_available=True)
        self.assertTrue(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.CLOSED.value)
        task_assignment.refresh_from_db()
        self.assertEqual(task_assignment.worker, worker2)

    def test_handle_staffing_response_not_is_available(self):
        # Test StaffingResponse object creation
        old_count = StaffingResponse.objects.all().count()
        request = self.staffing_request_inquiry.request
        new_request_inquiry = StaffingRequestInquiryFactory(
            request=request)
        response = handle_staffing_response(
            self.worker, self.staffing_request_inquiry.id,
            is_available=False)
        self.assertFalse(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.SENDING_INQUIRIES.value)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count + 1)

        # Replay of same request
        response = handle_staffing_response(
            self.worker, self.staffing_request_inquiry.id,
            is_available=False)
        self.assertFalse(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.SENDING_INQUIRIES.value)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count + 1)

        # Change mind to `is_available=True`
        response = handle_staffing_response(
            self.worker, self.staffing_request_inquiry.id,
            is_available=True)
        self.assertTrue(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.CLOSED.value)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count + 1)

        # Task is not available to claim
        new_request_inquiry = StaffingRequestInquiryFactory(
            request=self.staffing_request_inquiry.request,
            request__task__step__is_human=True)
        new_worker = new_request_inquiry.communication_preference.worker
        old_count = StaffingResponse.objects.all().count()
        response = handle_staffing_response(
            new_worker, new_request_inquiry.id, is_available=True)
        self.assertFalse(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.CLOSED.value)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count + 1)

    @patch('orchestra.communication.staffing.message_experts_slack_group')
    def test_address_staffing_requests(self, mock_slack):
        worker2 = WorkerFactory()
        CommunicationPreferenceFactory(
            worker=worker2,
            communication_type=(
                CommunicationPreference.CommunicationType
                .NEW_TASK_AVAILABLE.value))
        WorkerCertificationFactory(
            worker=worker2,
            certification=self.certification
        )

        request = StaffBotRequestFactory()
        request.task.step.required_certifications.add(self.certification)

        self.assertEqual(request.status,
                         StaffBotRequest.Status.SENDING_INQUIRIES.value)

        address_staffing_requests(worker_batch_size=1,
                                  frequency=timedelta(minutes=0))
        mock_slack.assert_not_called()
        request.refresh_from_db()
        self.assertEqual(request.status,
                         StaffBotRequest.Status.SENDING_INQUIRIES.value)
        # Inquiries increase by two because we send a Slack and an
        # email notification.
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            2)

        address_staffing_requests(worker_batch_size=1,
                                  frequency=timedelta(minutes=0))
        self.assertEqual(mock_slack.call_count, 1)
        mock_slack.reset()

        request.refresh_from_db()
        self.assertEqual(request.status,
                         StaffBotRequest.Status.SENDING_INQUIRIES.value)
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            4)

        # marked as closed and no new request inquiries sent.
        address_staffing_requests(worker_batch_size=1,
                                  frequency=timedelta(minutes=0))
        self.assertTrue(mock_slack.called)
        request.refresh_from_db()
        self.assertEqual(request.status,
                         StaffBotRequest.Status.DONE_SENDING_INQUIRIES.value)
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            4)

    @patch('orchestra.communication.staffing.handle_staffing_response')
    @patch('orchestra.communication.staffing.message_experts_slack_group')
    def test_address_staffing_request_priorities(
            self, mock_slack, mock_handle):
        worker2, _ = self._create_worker(-1)
        worker3, communication_preference3 = self._create_worker(1)

        # Workers should be contacted in priority order.
        excluded = [self.worker]
        for worker in (worker3, worker2):
            address_staffing_requests(worker_batch_size=1,
                                      frequency=timedelta(minutes=0))
            inquiries = (
                StaffingRequestInquiry.objects
                .filter(request=self.staffing_request_inquiry.request)
                .exclude(communication_preference__worker__in=excluded))
            # One inquiry for each communication method
            self.assertEqual(inquiries.count(), 2)
            for inquiry in inquiries:
                self.assertEqual(worker.id,
                                 inquiry.communication_preference.worker.id)
            excluded.append(worker)

        def _set_hours_available(availability, hours):
            for day in ['mon', 'tues', 'wed', 'thurs', 'fri', 'sat', 'sun']:
                setattr(availability, 'hours_available_{}'.format(day), hours)
            availability.save()

        total_inquiries = StaffingRequestInquiry.objects.count()
        # The highest-priority Worker has availability (7 hours) but
        # has already done 2 hours of work today and has been
        # previously assigned a 3-hour task today, so shouldn't be
        # automatically assigned the task which is estimated to take 3
        # hours.
        worker3_availability = WorkerAvailabilityFactory(worker=worker3)
        _set_hours_available(worker3_availability, 7)
        worker3.max_autostaff_hours_per_day = 7
        worker3.save()
        TimeEntryFactory(worker=worker3,
                         date=timezone.now().date(),
                         time_worked=timedelta(hours=2))
        self._create_inquired_staffing_request(
            worker3, communication_preference3, True)
        total_inquiries += 1
        address_staffing_requests(worker_batch_size=1,
                                  frequency=timedelta(minutes=0))
        mock_handle.assert_not_called()
        self.assertEqual(
            total_inquiries, StaffingRequestInquiry.objects.count())

        # Despite wanting 9 hours of work, and the estimate of worked
        # hours and new hours being 8, their maximum allowed hours
        # aren't high enough, so they aren't assigned.
        _set_hours_available(worker3_availability, 9)
        address_staffing_requests(worker_batch_size=1,
                                  frequency=timedelta(minutes=0))
        mock_handle.assert_not_called()
        self.assertEqual(
            total_inquiries, StaffingRequestInquiry.objects.count())

        # The highest-priority Worker has availability! But they have
        # already been assigned a task, and the maximum tasks allowed
        # is 1, so they aren't assigned.
        with override_settings(ORCHESTRA_MAX_AUTOSTAFF_TASKS_PER_DAY=1):
            worker3.max_autostaff_hours_per_day = 9
            worker3.save()
            address_staffing_requests(worker_batch_size=1,
                                      frequency=timedelta(minutes=0))
            mock_handle.assert_not_called()
            self.assertEqual(
                total_inquiries, StaffingRequestInquiry.objects.count())

        # The stars have aligned! Our default max tasks is 4, which
        # means this worker can pick up the tasks!
        address_staffing_requests(worker_batch_size=1,
                                  frequency=timedelta(minutes=0))
        previously_opted_in_method = (
            StaffingRequestInquiry.CommunicationMethod
            .PREVIOUSLY_OPTED_IN.value)
        inquiries = StaffingRequestInquiry.objects.filter(
            communication_preference__worker=worker3,
            communication_method=previously_opted_in_method
        )
        self.assertEqual(inquiries.count(), 1)
        mock_handle.assert_called_once_with(
            worker3, inquiries.first().id, is_available=True)
        self.assertEqual(
            total_inquiries + 1, StaffingRequestInquiry.objects.count())

    @patch('orchestra.communication.staffing.message_experts_slack_group')
    def test_handle_staffing_response_all_rejected(self, mock_slack):
        worker2 = WorkerFactory()

        CommunicationPreferenceFactory(
            worker=worker2,
            communication_type=(
                CommunicationPreference.CommunicationType
                .NEW_TASK_AVAILABLE.value))

        response = handle_staffing_response(
            self.worker, self.staffing_request_inquiry.id,
            is_available=False)
        self.assertFalse(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.CLOSED.value)
        self.assertEqual(mock_slack.call_count, 1)
        mock_slack.reset()

        handle_staffing_response(
            worker2, self.staffing_request_inquiry.id,
            is_available=False)

        self.assertEqual(mock_slack.call_count, 1)

    @patch('orchestra.communication.staffing.message_experts_slack_group')
    def test_get_available_request(self, mock_slack):
        # Close all open requests so new worker doesn't receive them.
        address_staffing_requests(worker_batch_size=2,
                                  frequency=timedelta(minutes=0))

        self.assertEqual(len(get_available_requests(self.worker)), 1)

        worker2 = WorkerFactory()
        CommunicationPreferenceFactory(
            worker=worker2,
            communication_type=(
                CommunicationPreference.CommunicationType
                .NEW_TASK_AVAILABLE.value))
        WorkerCertificationFactory(
            worker=worker2,
            certification=self.certification
        )

        request1 = StaffBotRequestFactory(task__step__is_human=True)
        request1.task.step.required_certifications.add(self.certification)
        request2 = StaffBotRequestFactory(task__step__is_human=True)
        request2.task.step.required_certifications.add(self.certification)

        address_staffing_requests(worker_batch_size=2,
                                  frequency=timedelta(minutes=0))
        inquiry1 = (
            StaffingRequestInquiry.objects
            .filter(communication_preference__worker=self.worker)
            .filter(request=request1).first())
        inquiry2 = (
            StaffingRequestInquiry.objects
            .filter(communication_preference__worker=worker2)
            .filter(request=request2).first())

        # `self.worker` now has three available tasks, whereas `worker2`
        # just has access to the two new tasks.
        available_requests = get_available_requests(self.worker)
        self.assertEqual(len(available_requests), 3)
        self.assertEqual(len(get_available_requests(worker2)), 2)

        # Tasks should be sorted by start_datetime in ascending order.

        first_available, second_available, third_available = (
            available_requests[0]['available_datetime'],
            available_requests[1]['available_datetime'],
            available_requests[2]['available_datetime'])
        self.assertLess(first_available, second_available)
        self.assertLess(second_available, third_available)

        # `self.worker` will lose an available task (they accept it),
        # whereas `worker2` is unchanged.
        handle_staffing_response(
            self.worker, self.staffing_request_inquiry.id,
            is_available=True)
        self.assertEqual(len(get_available_requests(self.worker)), 2)
        self.assertEqual(len(get_available_requests(worker2)), 2)

        # `self.worker` will lose an available task (they ignore it),
        # whereas `worker2` is unchanged.
        handle_staffing_response(
            self.worker, inquiry1.id,
            is_available=False)
        self.assertEqual(len(get_available_requests(self.worker)), 1)
        self.assertEqual(len(get_available_requests(worker2)), 2)

        # `worker2` takes a task.
        handle_staffing_response(
            worker2, inquiry2.id,
            is_available=True)
        self.assertEqual(len(get_available_requests(self.worker)), 0)
        self.assertEqual(len(get_available_requests(worker2)), 1)

    @patch('orchestra.communication.staffing.message_internal_slack_group')
    def test_warn_staffing_team_about_unstaffed_tasks(self, mock_slack):
        warn_staffing_team_about_unstaffed_tasks()
        mock_slack.assert_not_called()

        create_time = timezone.now() - timedelta(minutes=31)
        StaffingResponseFactory(
            request_inquiry__request__task__start_datetime=create_time,
            request_inquiry__request__created_at=create_time,
            is_winner=False
        )
        warn_staffing_team_about_unstaffed_tasks()
        self.assertEqual(mock_slack.call_count, 1)

        args, _ = mock_slack.call_args
        self.assertTrue('No winner request for task' in args[1])

    @patch('orchestra.communication.staffing.StaffBot'
           '._send_staffing_request_by_mail')
    @patch('orchestra.communication.staffing.StaffBot'
           '._send_staffing_request_by_slack')
    def test_remind_workers_about_available_tasks_has_winner(
            self, mock_slack, mock_mail):
        # mark existing request as a winner
        StaffingResponse.objects.create(
            request_inquiry=self.staffing_request_inquiry,
            is_available=True,
            is_winner=True)

        remind_workers_about_available_tasks()
        mock_slack.assert_not_called()

    @patch('orchestra.communication.staffing.StaffBot'
           '._send_staffing_request_by_mail')
    @patch('orchestra.communication.staffing.StaffBot'
           '._send_staffing_request_by_slack')
    def test_remind_workers_about_available_tasks(self, mock_slack, mock_mail):
        # mark existing request as a winner
        remind_workers_about_available_tasks()
        self.assertEqual(mock_slack.call_count, 1)

        args, _ = mock_slack.call_args
        self.assertTrue(
            'Tasks are still available for you to work on' in args[1])
        # No href in slack
        self.assertTrue(
            '<a href' not in args[1])

        args, _ = mock_mail.call_args
        self.assertTrue(
            'Tasks are still available for you to work on' in args[1])
        # href in email
        self.assertTrue(
            '<a href' in args[1])

    def test_mark_worker_as_winner(self):
        self.assertEqual(
            self.staffing_request_inquiry.responses.all().count(), 0)

        mark_worker_as_winner(self.worker,
                              self.staffing_request_inquiry.request.task,
                              0,
                              self.staffing_request_inquiry)
        staffbot_request = self.staffing_request_inquiry.request
        staffbot_request.refresh_from_db()

        self.assertEqual(staffbot_request.status,
                         StaffBotRequest.Status.CLOSED.value)
        self.assertEqual(
            self.staffing_request_inquiry.responses.all().count(), 0)

        # Can't mark another worker as a winner for this request
        worker2 = WorkerFactory()
        inquiry2 = StaffingRequestInquiryFactory(
            communication_preference__worker=worker2,
            communication_preference__communication_type=(
                CommunicationPreference.CommunicationType
                .NEW_TASK_AVAILABLE.value),
            request__task__step__is_human=True,
            request=self.staffing_request_inquiry.request
        )
        mark_worker_as_winner(worker2,
                              inquiry2.request.task,
                              0,
                              inquiry2)
        self.assertEqual(inquiry2.responses.count(), 0)

    @patch('orchestra.communication.staffing.message_experts_slack_group')
    def test_address_staffing_requests_parameters(self, mock_slack):
        for idx in range(5):
            worker = WorkerFactory()
            CommunicationPreferenceFactory(
                worker=worker,
                communication_type=(
                    CommunicationPreference.CommunicationType
                    .NEW_TASK_AVAILABLE.value))
            WorkerCertificationFactory(
                worker=worker,
                certification=self.certification
            )
        # Make a new request and turn off the global request, as the
        # global one has an inquiry on it already.
        request = StaffBotRequestFactory(task__step__is_human=True)
        request.task.step.required_certifications.add(self.certification)

        self.staffing_request_inquiry.request.status = (
            StaffBotRequest.Status.DONE_SENDING_INQUIRIES)

        address_staffing_requests(worker_batch_size=1,
                                  frequency=timedelta(minutes=20))
        # Inquiries increase by two because we send a Slack and an
        # email notification.  `last_inquiry_sent` is None on new
        # tasks, so we send this batch regardless of frequency.
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            2)

        address_staffing_requests(worker_batch_size=2,
                                  frequency=timedelta(minutes=20))
        # Don't send more inquiries, since it hasn't been 20 minutes.
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            2)

        request.last_inquiry_sent = timezone.now() - timedelta(minutes=21)
        request.save()
        address_staffing_requests(worker_batch_size=2,
                                  frequency=timedelta(minutes=20))
        # Send two email and two slack inquiries since it's been 21 minutes.
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            6)

        address_staffing_requests(worker_batch_size=10,
                                  frequency=timedelta(minutes=20))
        # Don't send more inquiries, since it hasn't been 20 minutes.
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            6)

        request.last_inquiry_sent = timezone.now() - timedelta(minutes=21)
        request.save()
        address_staffing_requests(worker_batch_size=10,
                                  frequency=timedelta(minutes=20))
        # Send remaining inquiries, since enough time has passed.
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            12)

        request.last_inquiry_sent = timezone.now() - timedelta(minutes=21)
        request.save()
        address_staffing_requests(worker_batch_size=10,
                                  frequency=timedelta(minutes=20))
        # We're all out of workers to whom we'd like to send inquiries.
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            12)



# Generated by Django 2.2.13 on 2021-07-26 09:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0095_todo_slug'),
    ]

    operations = [
        migrations.AddField(
            model_name='todo',
            name='required',
            field=models.BooleanField(default=False),
        ),
    ]



import logging
from urllib.parse import urlparse
from urllib.parse import urlunsplit

from django.urls import reverse
from jsonview.exceptions import BadRequest
from rest_framework import generics

from orchestra.core.errors import TaskAssignmentError
from orchestra.core.errors import WorkerCertificationError
from orchestra.models import Project
from orchestra.models import WorkerCertification
from orchestra.models import Workflow
from orchestra.models import Todo
from orchestra.models import TodoListTemplate
from orchestra.project import create_project_with_tasks
from orchestra.project_api.api import get_project_information
from orchestra.utils.decorators import api_endpoint
from orchestra.utils.load_json import load_encoded_json
from orchestra.utils.task_lifecycle import assign_task
from orchestra.utils.notifications import message_experts_slack_group
from orchestra.project_api.auth import OrchestraProjectAPIAuthentication
from orchestra.project_api.auth import IsSignedUser
from orchestra.todos.views import GenericTodoViewset
from orchestra.todos.serializers import TodoListTemplateSerializer
from orchestra.todos.serializers import BulkTodoSerializerWithoutQA
from orchestra.todos.api import add_todolist_template

logger = logging.getLogger(__name__)


@api_endpoint(methods=['POST'],
              permissions=(IsSignedUser,),
              logger=logger,
              auths=(OrchestraProjectAPIAuthentication,))
def project_information(request):
    try:
        data = load_encoded_json(request.body)
        project_ids = data['project_ids']
        return get_project_information(project_ids)
    except KeyError:
        raise BadRequest('project_ids is required')


@api_endpoint(methods=['POST'],
              permissions=(IsSignedUser,),
              logger=logger,
              auths=(OrchestraProjectAPIAuthentication,))
def create_project(request):
    project_details = load_encoded_json(request.body)
    try:
        if project_details['task_class'] == 'real':
            task_class = WorkerCertification.TaskClass.REAL
        else:
            task_class = WorkerCertification.TaskClass.TRAINING
        args = (
            project_details['workflow_slug'],
            project_details['workflow_version_slug'],
            project_details['description'],
            project_details['priority'],
            task_class,
            project_details['project_data'],
        )
    except KeyError:
        raise BadRequest('One of the parameters is missing')

    project = create_project_with_tasks(*args)
    return {'project_id': project.id}


@api_endpoint(methods=['POST'],
              permissions=(IsSignedUser,),
              logger=logger,
              auths=(OrchestraProjectAPIAuthentication,))
def project_details_url(request):
    project_details = load_encoded_json(request.body)
    project_id = project_details.get('project_id')

    if project_id is None:
        raise BadRequest('project_id parameter is missing')
    project_details_url = '{}/project/{}'.format(
        reverse('orchestra:index'), project_id)
    parsed_url = urlparse(request.build_absolute_uri())
    url = urlunsplit((parsed_url.scheme,
                      parsed_url.netloc,
                      project_details_url,
                      '',
                      ''))
    return {'project_details_url': url}


@api_endpoint(methods=['GET'],
              permissions=(IsSignedUser,),
              logger=logger,
              auths=(OrchestraProjectAPIAuthentication,))
def workflow_types(request):
    workflows = {
        w.slug: {
            'name': w.name,
            'versions': {
                v.slug: {
                    'name': v.name,
                    'description': v.description
                }
                for v in w.versions.all()
            }
        }
        for w in Workflow.objects.all()
    }
    return {'workflows': workflows}


@api_endpoint(methods=['POST'],
              permissions=(IsSignedUser,),
              logger=logger,
              auths=(OrchestraProjectAPIAuthentication,))
def assign_worker_to_task(request):
    data = load_encoded_json(request.body)
    errors = {}
    try:
        assign_task(data.get('worker_id'), data.get('task_id'))
    except WorkerCertificationError as e:
        errors['worker_certification_error'] = str(e)
    except TaskAssignmentError as e:
        errors['task_assignment_error'] = str(e)
    except Exception as e:
        errors['error'] = str(e)
    success = len(errors) == 0
    return {
        'success': success,
        'errors': errors,
    }


@api_endpoint(methods=['POST'],
              permissions=(IsSignedUser,),
              logger=logger,
              auths=(OrchestraProjectAPIAuthentication,))
def message_project_team(request):
    """
    Endpoint for sending arbitrary message to a project team.
    Payload example:
    {'message': 'Chat message', 'project_id': 'some-id-123'}
    """
    data = load_encoded_json(request.body)
    try:
        message = data['message']
        project_id = data['project_id']
        project = Project.objects.get(id=project_id)
    except KeyError:
        text = ('An object with `message` and `project_id` attributes'
                ' should be supplied')
        raise BadRequest(text)
    except Project.DoesNotExist:
        raise BadRequest('No project for given id')
    if project.slack_group_id:
        message_experts_slack_group(project.slack_group_id, message)
    else:
        error_message = (
            "The following project doesn't have slack_group_id: {}"
        ).format(project)
        raise BadRequest(error_message)
    return {'success': True}


@api_endpoint(methods=['POST'],
              permissions=(IsSignedUser,),
              logger=logger,
              auths=(OrchestraProjectAPIAuthentication,))
def create_todos_from_template(request):
    """
    Endpoint for creating todos in a project.
    Payload example:
    {
        'todolist_template_slug': 'some-template-slug-123',
        'step_slug': 'some-step-slug-123',
        'project_id': 'some-project-id-123'
        'additional_data': {
            'some_key': 'some_value'
        }
    }
    """
    data = load_encoded_json(request.body)
    try:
        todolist_template_slug = data.get('todolist_template_slug')
        step_slug = data.get('step_slug')
        project_id = data.get('project_id')
        additional_data = data.get('additional_data')
        if step_slug and project_id and todolist_template_slug:
            add_todolist_template(todolist_template_slug, project_id,
                                  step_slug, additional_data)
            todos = Todo.objects.filter(
                template__slug=todolist_template_slug,
                project__id=project_id,
                step__slug=step_slug).order_by('-created_at')
            serializer = BulkTodoSerializerWithoutQA(todos, many=True)
            return {
                'success': True,
                'todos': serializer.data
            }
        else:
            text = ('An object with `template_slug`, `step_slug`,'
                    ' and `project_id` attributes should be supplied')
            raise Exception(text)
    except Exception as e:
        return {
            'success': False,
            'errors': {
                'error': str(e)
            }
        }


class TodoApiViewset(GenericTodoViewset):
    """
    This viewset inherits from GenericTodoViewset and used by
    an orchestra-client facing endpoint, exposed via a router in api_urls.py
    """
    permission_classes = (IsSignedUser,)
    authentication_classes = (OrchestraProjectAPIAuthentication,)


class TodoTemplatesList(generics.ListAPIView):
    permission_classes = (IsSignedUser,)
    authentication_classes = (OrchestraProjectAPIAuthentication,)

    serializer_class = TodoListTemplateSerializer
    queryset = TodoListTemplate.objects.all()
